#
#   Copyright 2022 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#


from hsfs import user


class TestUser:
    def test_from_response_json(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["user"]["get"]["response"]

        # Act
        u = user.User.from_response_json(json)

        # Assert
        assert u.username == "test_username"
        assert u.email == "test_email"
        assert u.first_name == "test_first_name"
        assert u.last_name == "test_last_name"
        assert u.status == "test_status"
        assert u.secret == "test_secret"
        assert u.chosen_password == "test_chosen_password"
        assert u.repeated_password == "test_repeated_password"
        assert u.tos == "test_tos"
        assert u.two_factor == "test_two_factor"
        assert u.tours_state == "test_tours_state"
        assert u.max_num_projects == "test_max_num_projects"
        assert u.test_user == "test_test_user"
        assert u.user_account_type == "test_user_account_type"
        assert u.num_active_projects == "test_num_active_projects"
        assert u.num_remaining_projects == "test_num_remaining_projects"

    def test_from_response_json_basic_info(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["user"]["get_basic_info"]["response"]

        # Act
        u = user.User.from_response_json(json)

        # Assert
        assert u is None



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

from typing import Any, Dict, Optional

import humps
from hsfs import feature_group as feature_group_module


class HudiFeatureGroupAlias:
    def __init__(
        self,
        feature_group: Dict[str, Any],
        alias: str,
        left_feature_group_end_timestamp: Optional[int] = None,
        left_feature_group_start_timestamp: Optional[int] = None,
        **kwargs,
    ) -> None:
        self._feature_group = feature_group_module.FeatureGroup.from_response_json(
            feature_group
        )
        self._alias = alias
        self._left_feature_group_start_timestamp = left_feature_group_start_timestamp
        self._left_feature_group_end_timestamp = left_feature_group_end_timestamp

    @classmethod
    def from_response_json(cls, json_dict: Dict[str, Any]) -> HudiFeatureGroupAlias:
        json_decamelized = humps.decamelize(json_dict)
        return cls(**json_decamelized)

    @property
    def feature_group(self) -> "feature_group_module.FeatureGroup":
        return self._feature_group

    @property
    def alias(self) -> str:
        return self._alias

    @property
    def left_feature_group_start_timestamp(self) -> Optional[int]:
        return self._left_feature_group_start_timestamp

    @property
    def left_feature_group_end_timestamp(self) -> Optional[int]:
        return self._left_feature_group_end_timestamp



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

from typing import Dict, List, Optional, Union

from hsfs import client
from hsfs.core.feature_monitoring_result import FeatureMonitoringResult


class FeatureMonitoringResultApi:
    """Feature Monitoring Result endpoints for the Feature Group resource.

    :param feature_store_id: id of the respective Feature Store
    :type feature_store_id: int
    :param feature_group_id: id of the feature group, if attaching a config to a feature group
    :type feature_group_id: int, optional
    :param feature_view_name: name of the feature view, if attaching a config to a feature view
    :type feature_view_name: str, optional
    :param feature_view_version: version of the feature view, if attaching a config to a feature view
    :type feature_view_version: int, optional
    """

    def __init__(
        self,
        feature_store_id: int,
        feature_group_id: Optional[int] = None,
        feature_view_name: Optional[str] = None,
        feature_view_version: Optional[int] = None,
        **kwargs,
    ):
        if feature_group_id is None:
            assert feature_view_name is not None
            assert feature_view_version is not None

        self._feature_store_id = feature_store_id
        self._feature_group_id = feature_group_id
        self._feature_view_name = feature_view_name
        self._feature_view_version = feature_view_version

    def create(
        self,
        fm_result: FeatureMonitoringResult,
    ) -> FeatureMonitoringResult:
        """Create an feature monitoring result attached to the Feature of a Feature Group.

        :param fm_result: feature monitoring result object to be attached to a Feature
        :type fm_result: `FeatureMonitoringResult`
        :return: the created feature monitoring result
        :rtype: FeatureMonitoringResult
        """
        _client = client.get_instance()
        path_params = self.build_path_params(
            project_id=_client._project_id,
        )

        headers = {"content-type": "application/json"}
        payload = fm_result.json()
        return FeatureMonitoringResult.from_response_json(
            _client._send_request("POST", path_params, headers=headers, data=payload)
        )

    def delete(
        self,
        result_id: int,
    ) -> None:
        """Delete the Feature Monitoring result attached to a Feature."""
        _client = client.get_instance()
        path_params = self.build_path_params(
            project_id=_client._project_id,
        )
        path_params.append(result_id)

        _client._send_request("DELETE", path_params)

    def get_by_config_id(
        self,
        config_id: int,
        query_params: Optional[Dict[str, Union[str, List[str]]]] = None,
    ) -> List[FeatureMonitoringResult]:
        """Get all Feature Monitoring Result attached to a Monitoring Config.

        :param config_id: Id of the feature monitoring config for which to fetch all results
        :type config_id: int
        :param query_params: query parameters to filter the results by monitoring_time
        :type query_params: Dict[str, Union[str, List[str]]], optional
        :return: fetched feature monitoring results attached to the Feature Group
        :rtype: List[FeatureMonitoringResult]
        """
        _client = client.get_instance()
        path_params = self.build_path_params(
            project_id=_client._project_id,
        )
        path_params.append("byconfig")
        path_params.append(config_id)
        headers = {"content-type": "application/json"}

        return FeatureMonitoringResult.from_response_json(
            _client._send_request("GET", path_params, query_params, headers=headers)
        )

    def get_by_id(
        self,
        result_id: int,
    ) -> List[FeatureMonitoringResult]:
        """Get the Feature Monitoring Result attached to a Feature.

        :param result_id: Id of the feature monitoring result to fetch
        :type result_id: int
        :return: fetched feature monitoring result attached to the Feature Group
        :rtype: FeatureMonitoringResult || None
        """
        _client = client.get_instance()
        path_params = self.build_path_params(
            project_id=_client._project_id,
        )
        path_params.append(result_id)

        return FeatureMonitoringResult.from_response_json(
            _client._send_request("GET", path_params)
        )

    def build_path_params(
        self,
        project_id: int,
    ) -> List[str]:
        """Build the path parameters for the Feature Monitoring Result API.

        :param project_id: Id of the project
        :type project_id: int
        :return: list of path parameters
        :rtype: List[str]
        """
        path_params = [
            "project",
            project_id,
            "featurestores",
            self._feature_store_id,
        ]
        if self._feature_group_id is not None:
            path_params.extend(["featuregroups", self._feature_group_id])
        else:
            path_params.extend(
                [
                    "featureview",
                    self._feature_view_name,
                    "version",
                    self._feature_view_version,
                ]
            )
        path_params.extend(["featuremonitoring", "result"])

        return path_params



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import json
from typing import Mapping, Optional, Union

import humps
from hsfs import util


class FeatureDescriptiveStatistics:
    _SINGLE_VALUE_STATISTICS = [
        "count",
        "completeness",
        "num_non_null_values",
        "num_null_values",
        "approx_num_distinct_values",
        "min",
        "max",
        "sum",
        "mean",
        "stddev",
        "distinctness",
        "entropy",
        "uniqueness",
        "exact_num_distinct_values",
    ]

    def __init__(
        self,
        feature_name: str,
        feature_type: str = None,
        count: int = None,
        # for any feature type
        completeness: Optional[float] = None,
        num_non_null_values: Optional[int] = None,
        num_null_values: Optional[int] = None,
        approx_num_distinct_values: Optional[int] = None,
        # for numerical features
        min: Optional[float] = None,
        max: Optional[float] = None,
        sum: Optional[float] = None,
        mean: Optional[float] = None,
        stddev: Optional[float] = None,
        percentiles: Optional[Mapping[str, float]] = None,
        # with exact uniqueness
        distinctness: Optional[float] = None,
        entropy: Optional[float] = None,
        uniqueness: Optional[float] = None,
        exact_num_distinct_values: Optional[int] = None,
        extended_statistics: Optional[Union[dict, str]] = None,
        id: Optional[int] = None,
        **kwargs,
    ):
        self._id = id
        self._feature_type = feature_type
        self._feature_name = util.autofix_feature_name(feature_name)
        self._count = count
        self._completeness = completeness
        self._num_non_null_values = num_non_null_values
        self._num_null_values = num_null_values
        self._approx_num_distinct_values = approx_num_distinct_values
        self._min = min
        self._max = max
        self._sum = sum
        self._mean = mean
        self._stddev = stddev
        self._percentiles = percentiles
        self._distinctness = distinctness
        self._entropy = entropy
        self._uniqueness = uniqueness
        self._exact_num_distinct_values = exact_num_distinct_values
        self._extended_statistics = (
            extended_statistics
            if not isinstance(extended_statistics, str)
            else json.loads(extended_statistics)
        )

    def get_value(self, name):
        stat_name = name.lower()
        if stat_name not in self._SINGLE_VALUE_STATISTICS:
            return None
        try:
            return getattr(self, stat_name)
        except KeyError as err:
            raise AttributeError(f"'{name}' statistic has not been computed") from err

    @classmethod
    def from_response_json(cls, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        return cls(**json_decamelized)

    @classmethod
    def from_deequ_json(cls, json_dict: dict) -> "FeatureDescriptiveStatistics":
        stats_dict = {"feature_name": json_dict["column"]}

        if "dataType" in json_dict:
            stats_dict["feature_type"] = json_dict["dataType"]

        if "count" in json_dict and json_dict["count"] == 0:
            # if empty data, ignore the rest of statistics
            stats_dict["count"] = 0
            return cls(**stats_dict)

        # common for all data types
        if "numRecordsNull" in json_dict:
            stats_dict["num_null_values"] = json_dict["numRecordsNull"]
        if "numRecordsNonNull" in json_dict:
            stats_dict["num_non_null_values"] = json_dict["numRecordsNonNull"]
        if "numRecordsNull" in json_dict and "numRecordsNonNull" in json_dict:
            stats_dict["count"] = (
                json_dict["numRecordsNull"] + json_dict["numRecordsNonNull"]
            )
        if "count" in json_dict:
            stats_dict["count"] = json_dict["count"]
        if "completeness" in json_dict:
            stats_dict["completeness"] = json_dict["completeness"]
        if "approximateNumDistinctValues" in json_dict:
            stats_dict["approx_num_distinct_values"] = json_dict[
                "approximateNumDistinctValues"
            ]

        # commmon for all data types if exact_uniqueness is enabled
        if "uniqueness" in json_dict:
            stats_dict["uniqueness"] = json_dict["uniqueness"]
        if "entropy" in json_dict:
            stats_dict["entropy"] = json_dict["entropy"]
        if "distinctness" in json_dict:
            stats_dict["distinctness"] = json_dict["distinctness"]
        if "exactNumDistinctValues" in json_dict:
            stats_dict["exact_num_distinct_values"] = json_dict[
                "exactNumDistinctValues"
            ]

        # fractional / integral features
        if "minimum" in json_dict:
            stats_dict["min"] = json_dict["minimum"]
        if "maximum" in json_dict:
            stats_dict["max"] = json_dict["maximum"]
        if "sum" in json_dict:
            stats_dict["sum"] = json_dict["sum"]
        if "mean" in json_dict:
            stats_dict["mean"] = json_dict["mean"]
        if "stdDev" in json_dict:
            stats_dict["stddev"] = json_dict["stdDev"]
        if "approxPercentiles" in json_dict:
            stats_dict["percentiles"] = json_dict["approxPercentiles"]

        extended_statistics = {}
        if "unique_values" in json_dict:
            extended_statistics["unique_values"] = json_dict["unique_values"]
        if "correlations" in json_dict:
            extended_statistics["correlations"] = json_dict["correlations"]
        if "histogram" in json_dict:
            extended_statistics["histogram"] = json_dict["histogram"]
        if "kll" in json_dict:
            extended_statistics["kll"] = json_dict["kll"]
        stats_dict["extended_statistics"] = (
            extended_statistics if extended_statistics else None
        )

        return cls(**stats_dict)

    def to_dict(self):
        _dict = {
            "id": self._id,
            "featureName": self._feature_name,
            "count": self._count,
            "featureType": self._feature_type,
            "min": self._min,
            "max": self._max,
            "sum": self._sum,
            "mean": self._mean,
            "stddev": self._stddev,
            "completeness": self._completeness,
            "numNonNullValues": self._num_non_null_values,
            "numNullValues": self._num_null_values,
            "distinctness": self._distinctness,
            "entropy": self._entropy,
            "uniqueness": self._uniqueness,
            "approxNumDistinctValues": self._approx_num_distinct_values,
            "exactNumDistinctValues": self._exact_num_distinct_values,
            "percentiles": self._percentiles,
        }
        if self._extended_statistics is not None:
            _dict["extendedStatistics"] = json.dumps(self._extended_statistics)
        return _dict

    def json(self) -> str:
        return json.dumps(self, cls=util.FeatureStoreEncoder)

    def __str__(self):
        return self.json()

    def __repr__(self) -> str:
        return json.dumps(humps.decamelize(self.to_dict()), indent=2)

    @property
    def id(self) -> Optional[int]:
        return self._id

    @property
    def feature_type(self) -> str:
        """Data type of the feature. It can be one of Boolean, Fractional, Integral, or String."""
        return self._feature_type

    @property
    def feature_name(self) -> str:
        """Name of the feature."""
        return self._feature_name

    @property
    def count(self) -> int:
        """Number of values."""
        return self._count

    @property
    def completeness(self) -> Optional[float]:
        """Fraction of non-null values in a column."""
        return self._completeness

    @property
    def num_non_null_values(self) -> Optional[int]:
        """Number of non-null values."""
        return self._num_non_null_values

    @property
    def num_null_values(self) -> Optional[int]:
        """Number of null values."""
        return self._num_null_values

    @property
    def approx_num_distinct_values(self) -> Optional[int]:
        """Approximate number of distinct values."""
        return self._approx_num_distinct_values

    @property
    def min(self) -> Optional[float]:
        """Minimum value."""
        return self._min

    @property
    def max(self) -> Optional[float]:
        """Maximum value."""
        return self._max

    @property
    def sum(self) -> Optional[float]:
        """Sum of all feature values."""
        return self._sum

    @property
    def mean(self) -> Optional[float]:
        """Mean value."""
        return self._mean

    @property
    def stddev(self) -> Optional[float]:
        """Standard deviation of the feature values."""
        return self._stddev

    @property
    def percentiles(self) -> Optional[Mapping[str, float]]:
        """Percentiles."""
        return self._percentiles

    @property
    def distinctness(self) -> Optional[float]:
        """Fraction of distinct values of a feature over the number of all its values. Distinct values occur at least once.

        !!! note "Example"
            $[a, a, b]$ contains two distinct values $a$ and $b$, so distinctness is $2/3$.
        """
        return self._distinctness

    @property
    def entropy(self) -> Optional[float]:
        """Entropy is a measure of the level of information contained in an event (feature value) when considering all possible events (all feature values).
        Entropy is estimated using observed value counts as the negative sum of (value_count/total_count) * log(value_count/total_count).

        !!! note "Example"
            $[a, b, b, c, c]$ has three distinct values with counts $[1, 2, 2]$.

            Entropy is then $(-1/5*log(1/5)-2/5*log(2/5)-2/5*log(2/5)) = 1.055$.
        """
        return self._entropy

    @property
    def uniqueness(self) -> Optional[float]:
        """Fraction of unique values over the number of all values of a column. Unique values occur exactly once.

        !!! note "Example"
            $[a, a, b]$ contains one unique value $b$, so uniqueness is $1/3$.
        """
        return self._uniqueness

    @property
    def exact_num_distinct_values(self) -> Optional[int]:
        """Exact number of distinct values."""
        return self._exact_num_distinct_values

    @property
    def extended_statistics(self) -> Optional[dict]:
        """Additional statistics computed on the feature values such as histograms and correlations."""
        return self._extended_statistics



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

from typing import Literal, Optional, Union

from hsfs.client import external, hopsworks


_client = None


def init(
    client_type: Union[Literal["hopsworks"], Literal["external"]],
    host: Optional[str] = None,
    port: Optional[int] = None,
    project: Optional[str] = None,
    engine: Optional[str] = None,
    region_name: Optional[str] = None,
    secrets_store=None,
    hostname_verification: Optional[bool] = None,
    trust_store_path: Optional[str] = None,
    cert_folder: Optional[str] = None,
    api_key_file: Optional[str] = None,
    api_key_value: Optional[str] = None,
) -> None:
    global _client
    if not _client:
        if client_type == "hopsworks":
            _client = hopsworks.Client()
        elif client_type == "external":
            _client = external.Client(
                host,
                port,
                project,
                engine,
                region_name,
                secrets_store,
                hostname_verification,
                trust_store_path,
                cert_folder,
                api_key_file,
                api_key_value,
            )


def get_instance() -> Union[hopsworks.Client, external.Client]:
    global _client
    if _client:
        return _client
    raise Exception("Couldn't find client. Try reconnecting to Hopsworks.")


def stop() -> None:
    global _client
    _client._close()
    _client = None



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import warnings
from typing import List, Optional, Union

from hsfs import client, feature_group_commit, util
from hsfs import feature_group as fg_mod
from hsfs.core import explicit_provenance, ingestion_job, ingestion_job_conf, job


class FeatureGroupApi:
    BACKEND_FG_STREAM = "streamFeatureGroupDTO"
    BACKEND_FG_BATCH = "cachedFeaturegroupDTO"
    BACKEND_FG_EXTERNAL = "onDemandFeaturegroupDTO"
    BACKEND_FG_SPINE = "onDemandFeaturegroupDTO"

    def save(
        self,
        feature_group_instance: Union[
            fg_mod.FeatureGroup, fg_mod.ExternalFeatureGroup, fg_mod.SpineGroup
        ],
    ) -> Union[fg_mod.FeatureGroup, fg_mod.ExternalFeatureGroup, fg_mod.SpineGroup]:
        """Save feature group metadata to the feature store.

        :param feature_group_instance: metadata object of feature group to be
            saved
        :type feature_group_instance: FeatureGroup
        :return: updated metadata object of the feature group
        :rtype: FeatureGroup
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            feature_group_instance.feature_store_id,
            "featuregroups",
        ]
        headers = {"content-type": "application/json"}
        feature_group_object = feature_group_instance.update_from_response_json(
            _client._send_request(
                "POST",
                path_params,
                headers=headers,
                data=feature_group_instance.json(),
            ),
        )
        return feature_group_object

    def get(
        self, feature_store_id: int, name: str, version: Optional[int]
    ) -> Union[
        fg_mod.FeatureGroup,
        fg_mod.SpineGroup,
        fg_mod.ExternalFeatureGroup,
        List[fg_mod.FeatureGroup],
        List[fg_mod.SpineGroup],
        List[fg_mod.ExternalFeatureGroup],
    ]:
        """Get the metadata of a feature group with a certain name and version.

        :param feature_store_id: feature store id
        :type feature_store_id: int
        :param name: name of the feature group
        :type name: str
        :param version: version of the feature group
        :type version: int

        :return: feature group metadata object
        :rtype: FeatureGroup, SpineGroup, ExternalFeatureGroup, List[FeatureGroup], List[SpineGroup], List[ExternalFeatureGroup]
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            feature_store_id,
            "featuregroups",
            name,
        ]
        query_params = None if version is None else {"version": version}

        fg_objs = []
        # In principle unique names are enforced across fg type and this should therefore
        # return only list of the same type. But it cost nothing to check in case this gets forgotten.
        for fg_json in _client._send_request("GET", path_params, query_params):
            if (
                fg_json["type"] == FeatureGroupApi.BACKEND_FG_STREAM
                or fg_json["type"] == FeatureGroupApi.BACKEND_FG_BATCH
            ):
                fg_objs.append(fg_mod.FeatureGroup.from_response_json(fg_json))
            elif fg_json["type"] == FeatureGroupApi.BACKEND_FG_EXTERNAL:
                if fg_json.get("spine", False):
                    fg_objs.append(fg_mod.SpineGroup.from_response_json(fg_json))
                else:
                    fg_objs.append(
                        fg_mod.ExternalFeatureGroup.from_response_json(fg_json)
                    )
            else:
                list_of_types = [
                    FeatureGroupApi.BACKEND_FG_STREAM,
                    FeatureGroupApi.BACKEND_FG_BATCH,
                    FeatureGroupApi.BACKEND_FG_EXTERNAL,
                    FeatureGroupApi.BACKEND_FG_SPINE,
                ]
                raise ValueError(
                    f"Unknown feature group type: {fg_json['type']}, expected one of: "
                    + str(list_of_types)
                )

        if version is not None:
            self._check_features(fg_objs[0])
            return fg_objs[0]
        else:
            for fg_obj in fg_objs:
                self._check_features(fg_obj)
            return fg_objs

    def get_by_id(
        self, feature_store_id: int, feature_group_id: int
    ) -> Union[
        fg_mod.FeatureGroup,
        fg_mod.SpineGroup,
        fg_mod.ExternalFeatureGroup,
    ]:
        """Get the metadata of a feature group with a certain id.

        :param feature_store_id: feature store id
        :type feature_store_id: int
        :param feature_group_id: id of the feature group
        :type feature_group_id: int

        :return: feature group metadata object
        :rtype: FeatureGroup, SpineGroup, ExternalFeatureGroup
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            feature_store_id,
            "featuregroups",
            feature_group_id,
        ]

        fg_json = _client._send_request("GET", path_params)
        if (
            fg_json["type"] == FeatureGroupApi.BACKEND_FG_STREAM
            or fg_json["type"] == FeatureGroupApi.BACKEND_FG_BATCH
        ):
            return fg_mod.FeatureGroup.from_response_json(fg_json)
        elif fg_json["type"] == FeatureGroupApi.BACKEND_FG_EXTERNAL:
            if fg_json.get("spine", False):
                return fg_mod.SpineGroup.from_response_json(fg_json)
            else:
                return fg_mod.ExternalFeatureGroup.from_response_json(fg_json)

    def delete_content(
        self,
        feature_group_instance: Union[fg_mod.FeatureGroup, fg_mod.ExternalFeatureGroup],
    ) -> None:
        """Delete the content of a feature group.

        This endpoint serves to simulate the overwrite/insert mode.

        :param feature_group_instance: metadata object of feature group to clear
            the content for
        :type feature_group_instance: FeatureGroup
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            feature_group_instance.feature_store_id,
            "featuregroups",
            feature_group_instance.id,
            "clear",
        ]
        _client._send_request("POST", path_params)

    def delete(
        self,
        feature_group_instance: Union[
            fg_mod.FeatureGroup, fg_mod.ExternalFeatureGroup, fg_mod.SpineGroup
        ],
    ) -> None:
        """Drop a feature group from the feature store.

        Drops the metadata and data of a version of a feature group.

        :param feature_group_instance: metadata object of feature group
        :type feature_group_instance: FeatureGroup
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            feature_group_instance.feature_store_id,
            "featuregroups",
            feature_group_instance.id,
        ]
        _client._send_request("DELETE", path_params)

    def update_metadata(
        self,
        feature_group_instance: Union[
            fg_mod.FeatureGroup, fg_mod.ExternalFeatureGroup, fg_mod.SpineGroup
        ],
        feature_group_copy: Union[
            fg_mod.FeatureGroup, fg_mod.ExternalFeatureGroup, fg_mod.SpineGroup
        ],
        query_parameter: str,
        query_parameter_value=True,
    ) -> Union[fg_mod.FeatureGroup, fg_mod.ExternalFeatureGroup, fg_mod.SpineGroup]:
        """Update the metadata of a feature group.

        This only updates description and schema/features. The
        `feature_group_copy` is the metadata object sent to the backend, while
        `feature_group_instance` is the user object, which is only updated
        after a successful REST call.

        # Arguments
            feature_group_instance: FeatureGroup. User metadata object of the
                feature group.
            feature_group_copy: FeatureGroup. Metadata object of the feature
                group with the information to be updated.
            query_parameter: str. Query parameter that controls which information is updated. E.g. "updateMetadata".
            query_parameter_value: Str. Value of the query_parameter.

        # Returns
            FeatureGroup. The updated feature group metadata object.
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            feature_group_instance.feature_store_id,
            "featuregroups",
            feature_group_instance.id,
        ]
        headers = {"content-type": "application/json"}
        query_params = {query_parameter: query_parameter_value}
        feature_group_object = feature_group_instance.update_from_response_json(
            _client._send_request(
                "PUT",
                path_params,
                query_params,
                headers=headers,
                data=feature_group_copy.json(),
            ),
        )
        return feature_group_object

    def commit(
        self,
        feature_group_instance: fg_mod.FeatureGroup,
        feature_group_commit_instance: feature_group_commit.FeatureGroupCommit,
    ) -> feature_group_commit.FeatureGroupCommit:
        """
        Save feature group commit metadata.
        # Arguments
        feature_group_instance: FeatureGroup, required
            metadata object of feature group.
        feature_group_commit_instance: FeatureGroupCommit, required
            metadata object of feature group commit.
        # Returns
            `FeatureGroupCommit`.
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            feature_group_instance.feature_store_id,
            "featuregroups",
            feature_group_instance.id,
            "commits",
        ]
        headers = {"content-type": "application/json"}
        return feature_group_commit_instance.update_from_response_json(
            _client._send_request(
                "POST",
                path_params,
                headers=headers,
                data=feature_group_commit_instance.json(),
            ),
        )

    def get_commit_details(
        self,
        feature_group_instance: fg_mod.FeatureGroup,
        wallclock_timestamp: int,
        limit: int,
    ) -> feature_group_commit.FeatureGroupCommit:
        """
        Get feature group commit metadata.
        # Arguments
        feature_group_instance: FeatureGroup, required
            metadata object of feature group.
        limit: number of commits to retrieve
        wallclock_timestamp: specific point in time.
        # Returns
            `FeatureGroupCommit`.
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            feature_group_instance.feature_store_id,
            "featuregroups",
            feature_group_instance.id,
            "commits",
        ]
        headers = {"content-type": "application/json"}
        query_params = {"sort_by": "committed_on:desc", "offset": 0, "limit": limit}
        if wallclock_timestamp is not None:
            query_params["filter_by"] = "commited_on_ltoeq:" + str(wallclock_timestamp)

        return feature_group_commit.FeatureGroupCommit.from_response_json(
            _client._send_request("GET", path_params, query_params, headers=headers),
        )

    def ingestion(
        self,
        feature_group_instance: fg_mod.FeatureGroup,
        ingestion_conf: ingestion_job_conf.IngestionJobConf,
    ) -> ingestion_job.IngestionJob:
        """
        Setup a Hopsworks job for dataframe ingestion
        Args:
        feature_group_instance: FeatureGroup, required
            metadata object of feature group.
        ingestion_conf: the configuration for the ingestion job application
        """

        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            feature_group_instance.feature_store_id,
            "featuregroups",
            feature_group_instance.id,
            "ingestion",
        ]

        headers = {"content-type": "application/json"}
        return ingestion_job.IngestionJob.from_response_json(
            _client._send_request(
                "POST", path_params, headers=headers, data=ingestion_conf.json()
            ),
        )

    def update_table_schema(
        self,
        feature_group_instance: fg_mod.FeatureGroup,
    ) -> job.Job:
        """
        Setup a Hopsworks job to update table schema
        Args:
        feature_group_instance: FeatureGroup, required
            metadata object of feature group.
        job_conf: the configuration for the job application
        """

        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            feature_group_instance.feature_store_id,
            "featuregroups",
            feature_group_instance.id,
            "updatetableschema",
        ]

        headers = {"content-type": "application/json"}
        return job.Job.from_response_json(
            _client._send_request("POST", path_params, headers=headers),
        )

    def get_parent_feature_groups(
        self,
        feature_group_instance: Union[
            fg_mod.FeatureGroup, fg_mod.SpineGroup, fg_mod.ExternalFeatureGroup
        ],
    ) -> explicit_provenance.Links:
        """Get the parents of this feature group, based on explicit provenance.
        Parents are feature groups or external feature groups. These feature
        groups can be accessible, deleted or inaccessible.
        For deleted and inaccessible feature groups, only a minimal information is
        returned.

        # Arguments
            feature_group_instance: Metadata object of feature group.

        # Returns
            `ExplicitProvenance.Links`:  the feature groups used to generate this
            feature group
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            feature_group_instance.feature_store_id,
            "featuregroups",
            feature_group_instance.id,
            "provenance",
            "links",
        ]
        query_params = {
            "expand": "provenance_artifacts",
            "upstreamLvls": 1,
            "downstreamLvls": 0,
        }
        links_json = _client._send_request("GET", path_params, query_params)
        return explicit_provenance.Links.from_response_json(
            links_json,
            explicit_provenance.Links.Direction.UPSTREAM,
            explicit_provenance.Links.Type.FEATURE_GROUP,
        )

    def get_storage_connector_provenance(self, feature_group_instance):
        """Get the parents of this feature group, based on explicit provenance.
        Parents are storage connectors. These storage connector can be accessible,
        deleted or inaccessible.
        For deleted and inaccessible storage connector, only a minimal information is
        returned.

        # Arguments
            feature_group_instance: Metadata object of feature group.

        # Returns
            `ExplicitProvenance.Links`: the storage connector used to generated this
            feature group
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            feature_group_instance.feature_store_id,
            "featuregroups",
            feature_group_instance.id,
            "provenance",
            "links",
        ]
        query_params = {
            "expand": "provenance_artifacts",
            "upstreamLvls": 1,
            "downstreamLvls": 0,
        }
        links_json = _client._send_request("GET", path_params, query_params)
        return explicit_provenance.Links.from_response_json(
            links_json,
            explicit_provenance.Links.Direction.UPSTREAM,
            explicit_provenance.Links.Type.STORAGE_CONNECTOR,
        )

    def get_generated_feature_views(
        self,
        feature_group_instance: Union[
            fg_mod.FeatureGroup, fg_mod.SpineGroup, fg_mod.ExternalFeatureGroup
        ],
    ) -> explicit_provenance.Links:
        """Get the generated feature view using this feature group, based on explicit
        provenance. These feature views can be accessible or inaccessible. Explicit
        provenance does not track deleted generated feature view links, so deleted
        will always be empty.
        For inaccessible feature views, only a minimal information is returned.

        # Arguments
            feature_group_instance: Metadata object of feature group.

        # Returns
            `ExplicitProvenance.Links`: the feature views generated using this feature
            group
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            feature_group_instance.feature_store_id,
            "featuregroups",
            feature_group_instance.id,
            "provenance",
            "links",
        ]
        query_params = {
            "expand": "provenance_artifacts",
            "upstreamLvls": 0,
            "downstreamLvls": 1,
        }
        links_json = _client._send_request("GET", path_params, query_params)
        return explicit_provenance.Links.from_response_json(
            links_json,
            explicit_provenance.Links.Direction.DOWNSTREAM,
            explicit_provenance.Links.Type.FEATURE_VIEW,
        )

    def get_generated_feature_groups(
        self,
        feature_group_instance: Union[
            fg_mod.FeatureGroup, fg_mod.SpineGroup, fg_mod.ExternalFeatureGroup
        ],
    ) -> explicit_provenance.Links:
        """Get the generated feature groups using this feature group, based on explicit
        provenance. These feature groups can be accessible or inaccessible. Explicit
        provenance does not track deleted generated feature group links, so deleted
        will always be empty.
        For inaccessible feature groups, only a minimal information is returned.

        # Arguments
            feature_group_instance: Metadata object of feature group.

        # Returns
            `ExplicitProvenance.Links`: the feature groups generated using this
            feature group
        """
        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "featurestores",
            feature_group_instance.feature_store_id,
            "featuregroups",
            feature_group_instance.id,
            "provenance",
            "links",
        ]
        query_params = {
            "expand": "provenance_artifacts",
            "upstreamLvls": 0,
            "downstreamLvls": 1,
        }
        links_json = _client._send_request("GET", path_params, query_params)
        return explicit_provenance.Links.from_response_json(
            links_json,
            explicit_provenance.Links.Direction.DOWNSTREAM,
            explicit_provenance.Links.Type.FEATURE_GROUP,
        )

    def _check_features(self, feature_group_instance) -> None:
        if not feature_group_instance._features:
            warnings.warn(
                f"Feature Group `{feature_group_instance._name}`, version `{feature_group_instance._version}` has no features (to resolve this issue contact the admin or delete and recreate the feature group)",
                util.FeatureGroupWarning,
                stacklevel=1,
            )



#
#   Copyright 2021 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import datetime
from functools import partial
from typing import Dict, Optional, Union

import hsfs
import numpy
from hsfs import (
    feature_view,
    statistics,
    training_dataset,
    training_dataset_feature,
    transformation_function_attached,
    util,
)
from hsfs.core import (
    feature_view_api,
    statistics_api,
    statistics_engine,
    transformation_function_api,
)
from hsfs.core.builtin_transformation_function import BuiltInTransformationFunction


class TransformationFunctionEngine:
    BUILTIN_FN_NAMES = [
        "min_max_scaler",
        "standard_scaler",
        "robust_scaler",
        "label_encoder",
    ]
    AMBIGUOUS_FEATURE_ERROR = (
        "Provided feature '{}' in transformation functions is ambiguous and exists in more than one feature groups."
        "You can provide the feature with the prefix that was specified in the join."
    )
    FEATURE_NOT_EXIST_ERROR = "Provided feature '{}' in transformation functions do not exist in any of the feature groups."

    def __init__(self, feature_store_id: int):
        self._feature_store_id = feature_store_id
        self._transformation_function_api = (
            transformation_function_api.TransformationFunctionApi(feature_store_id)
        )
        self._statistics_api = statistics_api.StatisticsApi(
            feature_store_id, training_dataset.TrainingDataset.ENTITY_TYPE
        )
        self._feature_view_api: Optional["feature_view_api.FeatureViewApi"] = None
        self._statistics_engine: Optional["statistics_engine.StatisticsEngine"] = None

    def save(self, transformation_fn_instance):
        if self.is_builtin(transformation_fn_instance):
            raise ValueError(
                "Transformation function name '{name:}' with version 1 is reserved for built-in hsfs "
                "functions. Please use other name or version".format(
                    name=transformation_fn_instance.name
                )
            )
        if not callable(transformation_fn_instance.transformation_fn):
            raise ValueError("transformer must be callable")
        self._transformation_function_api.register_transformation_fn(
            transformation_fn_instance
        )

    def get_transformation_fn(self, name, version=None):
        transformation_fn_instances = (
            self._transformation_function_api.get_transformation_fn(name, version)
        )
        return transformation_fn_instances[0]

    def get_transformation_fns(self):
        transformation_fn_instances = (
            self._transformation_function_api.get_transformation_fn(
                name=None, version=None
            )
        )
        transformation_fns = []
        for (
            transformation_fn_instance
        ) in transformation_fn_instances:  # todo what is the point of this?
            transformation_fns.append(transformation_fn_instance)
        return transformation_fns

    def delete(self, transformation_function_instance):
        self._transformation_function_api.delete(transformation_function_instance)

    def get_td_transformation_fn(self, training_dataset):
        attached_transformation_fns = (
            self._transformation_function_api.get_td_transformation_fn(training_dataset)
        )
        transformation_fn_dict = {}
        for attached_transformation_fn in attached_transformation_fns:
            transformation_fn_dict[attached_transformation_fn.name] = (
                attached_transformation_fn.transformation_function
            )
        return transformation_fn_dict

    @staticmethod
    def attach_transformation_fn(training_dataset_obj=None, feature_view_obj=None):
        if training_dataset_obj:
            target_obj = training_dataset_obj  # todo why provide td and fv just to convert to target_obj?
        else:
            target_obj = feature_view_obj

        if target_obj._transformation_functions:
            for (
                feature_name,
                transformation_fn,
            ) in target_obj._transformation_functions.items():
                if feature_name in target_obj.labels:
                    raise ValueError(
                        "Online transformations for training dataset labels are not supported."
                    )

                feature, prefix, featuregroup = target_obj.query._get_feature_by_name(
                    feature_name
                )
                target_obj._features.append(
                    training_dataset_feature.TrainingDatasetFeature(
                        name=feature_name,
                        feature_group_feature_name=feature.name,
                        featuregroup=featuregroup,
                        type=transformation_fn.output_type,
                        label=False,
                        transformation_function=transformation_fn,
                    )
                )

    def is_builtin(self, transformation_fn_instance):
        return (
            transformation_fn_instance.name in self.BUILTIN_FN_NAMES
            and transformation_fn_instance.version == 1
        )

    @staticmethod
    def populate_builtin_fn_arguments(
        feature_name, transformation_function_instance, feature_descriptive_stats
    ):
        if transformation_function_instance.name == "min_max_scaler":
            min_value, max_value = BuiltInTransformationFunction.min_max_scaler_stats(
                feature_descriptive_stats, feature_name
            )
            transformation_function_instance.transformation_fn = partial(
                transformation_function_instance.transformation_fn,
                min_value=min_value,
                max_value=max_value,
            )
        elif transformation_function_instance.name == "standard_scaler":
            mean, std_dev = BuiltInTransformationFunction.standard_scaler_stats(
                feature_descriptive_stats, feature_name
            )
            transformation_function_instance.transformation_fn = partial(
                transformation_function_instance.transformation_fn,
                mean=mean,
                std_dev=std_dev,
            )
        elif transformation_function_instance.name == "robust_scaler":
            robust_scaler_stats = BuiltInTransformationFunction.robust_scaler_stats(
                feature_descriptive_stats, feature_name
            )
            transformation_function_instance.transformation_fn = partial(
                transformation_function_instance.transformation_fn,
                p25=robust_scaler_stats[24],
                p50=robust_scaler_stats[49],
                p75=robust_scaler_stats[74],
            )
        elif transformation_function_instance.name == "label_encoder":
            value_to_index = BuiltInTransformationFunction.encoder_stats(
                feature_descriptive_stats, feature_name
            )
            transformation_function_instance.transformation_fn = partial(
                transformation_function_instance.transformation_fn,
                value_to_index=value_to_index,
            )
        else:
            raise ValueError("Not implemented")

        return transformation_function_instance

    def populate_builtin_attached_fns(
        self, attached_transformation_fns, feature_descriptive_stats
    ):
        for ft_name in attached_transformation_fns:
            if self.is_builtin(attached_transformation_fns[ft_name]):
                # check if its built-in transformation function and populated with statistics arguments
                transformation_fn = self.populate_builtin_fn_arguments(
                    ft_name,
                    attached_transformation_fns[ft_name],
                    feature_descriptive_stats,
                )
                attached_transformation_fns[ft_name] = transformation_fn
        return attached_transformation_fns

    @staticmethod
    def infer_spark_type(output_type):
        if not output_type:
            return "STRING"  # STRING is default type for spark udfs

        if isinstance(output_type, str):
            if output_type.endswith("Type()"):
                return util.translate_legacy_spark_type(output_type)
            output_type = output_type.lower()

        if output_type in (str, "str", "string"):
            return "STRING"
        elif output_type in (bytes, "binary"):
            return "BINARY"
        elif output_type in (numpy.int8, "int8", "byte", "tinyint"):
            return "BYTE"
        elif output_type in (numpy.int16, "int16", "short", "smallint"):
            return "SHORT"
        elif output_type in (int, "int", "integer", numpy.int32):
            return "INT"
        elif output_type in (numpy.int64, "int64", "long", "bigint"):
            return "LONG"
        elif output_type in (float, "float"):
            return "FLOAT"
        elif output_type in (numpy.float64, "float64", "double"):
            return "DOUBLE"
        elif output_type in (
            datetime.datetime,
            numpy.datetime64,
            "datetime",
            "timestamp",
        ):
            return "TIMESTAMP"
        elif output_type in (datetime.date, "date"):
            return "DATE"
        elif output_type in (bool, "boolean", "bool"):
            return "BOOLEAN"
        else:
            raise TypeError("Not supported type %s." % output_type)

    @staticmethod
    def compute_transformation_fn_statistics(
        training_dataset_obj,
        builtin_tffn_features,
        label_encoder_features,
        feature_dataframe,
        feature_view_obj,
    ) -> statistics.Statistics:
        return training_dataset_obj._statistics_engine.compute_transformation_fn_statistics(
            td_metadata_instance=training_dataset_obj,
            columns=builtin_tffn_features,  # excluding label encoded features
            label_encoder_features=label_encoder_features,  # label encoded features only
            feature_dataframe=feature_dataframe,
            feature_view_obj=feature_view_obj,
        )

    @staticmethod
    def populate_builtin_transformation_functions(
        training_dataset, feature_view_obj, dataset
    ):
        # check if there any transformation functions that require statistics attached to td features
        builtin_tffn_label_encoder_features = [
            ft_name
            for ft_name in training_dataset.transformation_functions
            if training_dataset._transformation_function_engine.is_builtin(
                training_dataset.transformation_functions[ft_name]
            )
            and training_dataset.transformation_functions[ft_name].name
            == "label_encoder"
        ]
        builtin_tffn_features = [
            ft_name
            for ft_name in training_dataset.transformation_functions
            if training_dataset._transformation_function_engine.is_builtin(
                training_dataset.transformation_functions[ft_name]
            )
            and training_dataset.transformation_functions[ft_name].name
            != "label_encoder"
        ]

        if builtin_tffn_features or builtin_tffn_label_encoder_features:
            if training_dataset.splits:
                # compute statistics before transformations are applied
                stats = (
                    TransformationFunctionEngine.compute_transformation_fn_statistics(
                        training_dataset,
                        builtin_tffn_features,
                        builtin_tffn_label_encoder_features,
                        dataset.get(training_dataset.train_split),
                        feature_view_obj,
                    )
                )
            else:
                # compute statistics before transformations are applied
                stats = (
                    TransformationFunctionEngine.compute_transformation_fn_statistics(
                        training_dataset,
                        builtin_tffn_features,
                        builtin_tffn_label_encoder_features,
                        dataset,
                        feature_view_obj,
                    )
                )
            # Populate builtin transformations (if any) with respective arguments
            return training_dataset._transformation_function_engine.populate_builtin_attached_fns(
                training_dataset.transformation_functions,
                stats.feature_descriptive_statistics,
            )

    def get_ready_to_use_transformation_fns(
        self,
        entity: Union[hsfs.feature_view.FeatureView, training_dataset.TrainingDataset],
        training_dataset_version: Optional[int] = None,
    ) -> Dict[
        str, hsfs.transformation_function_attached.TransformationFunctionAttached
    ]:
        is_feat_view = isinstance(entity, feature_view.FeatureView)
        if self._feature_view_api is None:
            self._feature_view_api = feature_view_api.FeatureViewApi(
                self._feature_store_id
            )
        if self._statistics_engine is None:
            self._statistics_engine = statistics_engine.StatisticsEngine(
                self._feature_store_id,
                entity_type="featureview" if is_feat_view else "trainingdataset",
            )
        # get attached transformation functions
        transformation_functions = (
            self.get_td_transformation_fn(entity)
            if isinstance(entity, training_dataset.TrainingDataset)
            else (self.get_fv_attached_transformation_fn(entity.name, entity.version))
        )
        is_stat_required = (
            len(
                set(self.BUILTIN_FN_NAMES).intersection(
                    set([tf.name for tf in transformation_functions.values()])
                )
            )
            > 0
        )
        if not is_stat_required:
            td_tffn_stats = None
        else:
            # if there are any built-in transformation functions get related statistics and
            # populate with relevant arguments
            # there should be only one statistics object with before_transformation=true
            if is_feat_view and training_dataset_version is None:
                raise ValueError(
                    "Training data version is required for transformation. Call `feature_view.init_serving(version)` "
                    "or `feature_view.init_batch_scoring(version)` to pass the training dataset version."
                    "Training data can be created by `feature_view.create_training_data` or `feature_view.training_data`."
                )
            td_tffn_stats = self._statistics_engine.get(
                entity,
                before_transformation=True,
                training_dataset_version=training_dataset_version,
            )

        if is_stat_required and td_tffn_stats is None:
            raise ValueError(
                "No statistics available for initializing transformation functions."
                + "Training data can be created by `feature_view.create_training_data` or `feature_view.training_data`."
            )

        transformation_fns = self.populate_builtin_attached_fns(
            transformation_functions,
            td_tffn_stats.feature_descriptive_statistics
            if td_tffn_stats is not None
            else None,
        )
        return transformation_fns

    def get_fv_attached_transformation_fn(
        self, fv_name: str, fv_version: int
    ) -> Dict[str, "transformation_function_attached.TransformationFunctionAttached"]:
        if self._feature_view_api is None:
            self._feature_view_api = feature_view_api.FeatureViewApi(
                self._feature_store_id
            )
            self._statistics_engine = statistics_engine.StatisticsEngine(
                self._feature_store_id,
                entity_type="featureview",
            )
        transformation_functions = (
            self._feature_view_api.get_attached_transformation_fn(fv_name, fv_version)
        )
        if isinstance(transformation_functions, list):
            transformation_functions_dict = dict(
                [
                    (tf.name, tf.transformation_function)
                    for tf in transformation_functions
                ]
            )
        else:
            transformation_functions_dict = {
                transformation_functions.name: transformation_functions.transformation_function
            }
        return transformation_functions_dict



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import json
from typing import Any, Dict, List, Optional, Union

import humps
from hsfs import util
from hsfs.core.feature_descriptive_statistics import FeatureDescriptiveStatistics
from hsfs.split_statistics import SplitStatistics


class Statistics:
    DEFAULT_ROW_PERCENTAGE = 1.0

    def __init__(
        self,
        computation_time: int,
        row_percentage: float = 1.0,
        feature_descriptive_statistics: Optional[
            Union[
                FeatureDescriptiveStatistics,
                List[FeatureDescriptiveStatistics],
                Dict[str, Any],
            ]
        ] = None,
        # feature group
        feature_group_id: Optional[int] = None,
        window_start_commit_time: Optional[int] = None,
        window_end_commit_time: Optional[int] = None,
        # training dataset
        feature_view_name: Optional[str] = None,
        feature_view_version: Optional[int] = None,
        training_dataset_version: Optional[int] = None,
        split_statistics: Optional[
            Union[List[Dict[str, Any]], List[SplitStatistics]]
        ] = None,
        before_transformation: bool = False,
        href: Optional[str] = None,
        expand: Optional[str] = None,
        items: Optional[Dict[str, Any]] = None,
        count: Optional[int] = None,
        type: Optional[str] = None,
        **kwargs,
    ) -> None:
        self._computation_time = computation_time
        self._feature_descriptive_statistics = self._parse_descriptive_statistics(
            feature_descriptive_statistics
        )
        self._row_percentage = row_percentage
        # feature group
        self._feature_group_id = feature_group_id
        self._window_start_commit_time = window_start_commit_time
        self._window_end_commit_time = window_end_commit_time
        # training dataset
        self._feature_view_name = feature_view_name
        self._feature_view_version = feature_view_version
        self._training_dataset_version = training_dataset_version
        self._split_statistics = self._parse_split_statistics(split_statistics)
        self._before_transformation = before_transformation

    def _parse_descriptive_statistics(
        self,
        desc_statistics: Union[
            Dict[str, Any],
            FeatureDescriptiveStatistics,
            List[Dict[str, Any]],
            List[FeatureDescriptiveStatistics],
        ],
    ) -> Optional[List[FeatureDescriptiveStatistics]]:
        if desc_statistics is None:
            return None
        elif isinstance(desc_statistics, FeatureDescriptiveStatistics):
            return [desc_statistics]
        elif isinstance(desc_statistics, dict) and "items" not in desc_statistics:
            return [FeatureDescriptiveStatistics.from_response_json(desc_statistics)]
        elif isinstance(desc_statistics, dict) and "items" in desc_statistics:
            return [
                FeatureDescriptiveStatistics.from_response_json(fds)
                for fds in desc_statistics["items"]
            ]
        elif isinstance(desc_statistics, list):
            return [
                (
                    fds
                    if isinstance(fds, FeatureDescriptiveStatistics)
                    else FeatureDescriptiveStatistics.from_response_json(fds)
                )
                for fds in desc_statistics
            ]
        else:
            raise ValueError(
                "Descriptive statistics must be a FeatureDescriptiveStatistics object or a dictionary"
            )

    def _parse_split_statistics(
        self,
        split_statistics: Optional[Union[List[Dict[str, Any]], List[SplitStatistics]]],
    ) -> Optional[List[SplitStatistics]]:
        if split_statistics is None:
            return None
        return [
            (
                SplitStatistics.from_response_json(split)
                if isinstance(split, dict)
                else split
            )
            for split in split_statistics
        ]

    @classmethod
    def from_response_json(
        cls, json_dict: Dict[str, Any]
    ) -> Optional[Union["Statistics", List["Statistics"]]]:
        json_decamelized: dict = humps.decamelize(json_dict)
        # for consistency, if the json dict contains "count" and "items", we return a list
        # even when there is a single statistics in the list
        if "count" in json_decamelized:
            if json_decamelized["count"] == 0 or len(json_decamelized["items"]) == 0:
                return None
            else:
                return [cls(**config) for config in json_decamelized["items"]]
        else:
            return cls(**json_decamelized)

    def to_dict(self) -> Dict[str, Any]:
        # fg_id, fv_name, fv_version and td_version are already defined in the URI
        _dict = {
            "computationTime": self._computation_time,
            "rowPercentage": self._row_percentage,
            "windowStartCommitTime": self._window_start_commit_time,
            "windowEndCommitTime": self._window_end_commit_time,
            "beforeTransformation": self._before_transformation,
        }
        if self._feature_descriptive_statistics is not None:
            _dict["featureDescriptiveStatistics"] = [
                fds.to_dict() for fds in self._feature_descriptive_statistics
            ]
        if self._split_statistics is not None:
            _dict["splitStatistics"] = [sps.to_dict() for sps in self._split_statistics]
        return _dict

    def json(self) -> str:
        return json.dumps(self, cls=util.FeatureStoreEncoder)

    def __str__(self) -> str:
        return self.json()

    def __repr__(self) -> str:
        return f"Statistics({self._computation_time!r})"

    @property
    def computation_time(self) -> int:
        """Time at which the statistics were computed."""
        return self._computation_time

    @property
    def row_percentage(self) -> float:
        """Percentage of data on which statistics were computed."""
        return self._row_percentage

    @row_percentage.setter
    def row_percentage(self, row_percentage: Optional[float]):
        if isinstance(row_percentage, int) or isinstance(row_percentage, float):
            row_percentage = float(row_percentage)
            if row_percentage <= 0.0 or row_percentage > 1.0:
                raise ValueError("Row percentage must be a float between 0 and 1.")
            self._row_percentage = row_percentage
        elif row_percentage is None:
            self._row_percentage = self.DEFAULT_ROW_PERCENTAGE
        else:
            raise TypeError("Row percentage must be a float between 0 and 1.")

    @property
    def feature_descriptive_statistics(
        self,
    ) -> Optional[List[FeatureDescriptiveStatistics]]:
        """List of feature descriptive statistics."""
        return self._feature_descriptive_statistics

    @property
    def feature_group_id(self) -> Optional[int]:
        """Id of the feature group on whose data the statistics were computed."""
        return self._feature_group_id

    @property
    def feature_view_name(self) -> Optional[str]:
        """Name of the feature view whose query was used to retrieve the data on which the statistics were computed."""
        return self._feature_view_name

    @property
    def feature_view_version(self) -> Optional[int]:
        """Id of the feature view whose query was used to retrieve the data on which the statistics were computed."""
        return self._feature_view_version

    @property
    def window_start_commit_time(self) -> Optional[int]:
        """Start time of the window of data on which statistics were computed."""
        return self._window_start_commit_time

    @property
    def window_end_commit_time(self) -> Optional[int]:
        """End time of the window of data on which statistics were computed."""
        return self._window_end_commit_time

    @property
    def training_dataset_version(self) -> Optional[int]:
        """Version of the training dataset on which statistics were computed."""
        return self._training_dataset_version

    @property
    def split_statistics(self) -> Optional[List[SplitStatistics]]:
        """List of statistics computed on each split of a training dataset."""
        return self._split_statistics

    @property
    def before_transformation(self) -> bool:
        """Whether or not the statistics were computed on feature values before applying model-dependent transformations."""
        return self._before_transformation



#
#   Copyright 2022 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#


from hsfs import training_dataset_split


class TestTrainingDatasetSplit:
    def test_from_response_json(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["training_dataset_split"]["get"]["response"]

        # Act
        td_split = training_dataset_split.TrainingDatasetSplit.from_response_json(json)

        # Assert
        assert td_split.name == "test_name"
        assert td_split.percentage == "test_percentage"
        assert td_split.split_type == "test_split_type"
        assert td_split.start_time == "test_start_time"
        assert td_split.end_time == "test_end_time"

    def test_from_response_json_basic_info(self, backend_fixtures):
        # Arrange
        json = backend_fixtures["training_dataset_split"]["get_basic_info"]["response"]

        # Act
        td_split = training_dataset_split.TrainingDatasetSplit.from_response_json(json)

        # Assert
        assert td_split.name == "test_name"
        assert td_split.percentage is None
        assert td_split.split_type == "test_split_type"
        assert td_split.start_time is None
        assert td_split.end_time is None



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import asyncio
import json
import logging
import re
from typing import Any, Dict, List, Optional, Set, Tuple, Union

import aiomysql
import aiomysql.utils
from hsfs import feature_view, storage_connector, training_dataset, util
from hsfs.constructor.serving_prepared_statement import ServingPreparedStatement
from hsfs.core import feature_view_api, storage_connector_api, training_dataset_api
from hsfs.serving_key import ServingKey
from sqlalchemy import bindparam, exc, sql, text


_logger = logging.getLogger(__name__)


class OnlineStoreSqlClient:
    BATCH_HELPER_KEY = "batch_helper_column"
    SINGLE_HELPER_KEY = "single_helper_column"
    BATCH_VECTOR_KEY = "batch_feature_vectors"
    SINGLE_VECTOR_KEY = "single_feature_vector"

    def __init__(
        self,
        feature_store_id: id,
        skip_fg_ids: Optional[Set[int]],
        external: bool,
        serving_keys: Optional[Set[ServingKey]] = None,
        connection_options: Optional[Dict[str, Any]] = None,
    ):
        if _logger.isEnabledFor(logging.DEBUG):
            _logger.debug("Initialising Online Store Sql Client")
        self._feature_store_id = feature_store_id
        self._skip_fg_ids: Set[int] = skip_fg_ids or set()
        self._external = external

        self._prefix_by_serving_index = None
        self._pkname_by_serving_index = None
        self._serving_key_by_serving_index: Dict[str, ServingKey] = {}
        self._connection_pool = None
        self._serving_keys: Set[ServingKey] = set(serving_keys or [])

        self._prepared_statements: Dict[str, List[ServingPreparedStatement]] = {}
        self._parametrised_prepared_statements = {}
        self._prepared_statement_engine = None

        self._feature_view_api = feature_view_api.FeatureViewApi(feature_store_id)
        self._training_dataset_api = training_dataset_api.TrainingDatasetApi(
            feature_store_id
        )
        self._storage_connector_api = storage_connector_api.StorageConnectorApi()
        self._online_connector = None
        self._hostname = None
        self._connection_options = None

    def fetch_prepared_statements(
        self,
        entity: Union[feature_view.FeatureView, training_dataset.TrainingDataset],
        inference_helper_columns: bool,
    ) -> None:
        if isinstance(entity, feature_view.FeatureView):
            if _logger.isEnabledFor(logging.DEBUG):
                _logger.debug(
                    f"Initialising prepared statements for feature view {entity.name} version {entity.version}."
                )
            for key in self.get_prepared_statement_labels(inference_helper_columns):
                if _logger.isEnabledFor(logging.DEBUG):
                    _logger.debug(f"Fetching prepared statement for key {key}")
                self.prepared_statements[key] = (
                    self.feature_view_api.get_serving_prepared_statement(
                        entity.name,
                        entity.version,
                        batch=key.startswith("batch"),
                        inference_helper_columns=key.endswith("helper_column"),
                    )
                )
                if _logger.isEnabledFor(logging.DEBUG):
                    _logger.debug(f"{self.prepared_statements[key]}")
        elif isinstance(entity, training_dataset.TrainingDataset):
            if _logger.isEnabledFor(logging.DEBUG):
                _logger.debug(
                    f"Initialising prepared statements for training dataset {entity.name} version {entity.version}."
                )
            for key in self.get_prepared_statement_labels(
                with_inference_helper_column=False
            ):
                if _logger.isEnabledFor(logging.DEBUG):
                    _logger.debug(f"Fetching prepared statement for key {key}")
                self.prepared_statements[key] = (
                    self.training_dataset_api.get_serving_prepared_statement(
                        entity, batch=key.startswith("batch")
                    )
                )
        else:
            raise ValueError(
                "Object type needs to be `feature_view.FeatureView` or `training_dataset.TrainingDataset`."
            )

        if len(self.skip_fg_ids) > 0:
            if _logger.isEnabledFor(logging.DEBUG):
                _logger.debug(
                    f"Skip feature groups {self.skip_fg_ids} when initialising prepared statements."
                )
            self.prepared_statements[key] = {
                ps
                for ps in self.prepared_statements[key]
                if ps.feature_group_id not in self.skip_fg_ids
            }

    def init_prepared_statements(
        self,
        entity: Union[feature_view.FeatureView, training_dataset.TrainingDataset],
        inference_helper_columns: bool,
    ) -> None:
        if _logger.isEnabledFor(logging.DEBUG):
            _logger.debug(
                "Fetch and reset prepared statements and external as user may be re-initialising with different parameters"
            )
        self.fetch_prepared_statements(entity, inference_helper_columns)

        self.init_parametrize_and_serving_utils(
            self.prepared_statements[self.BATCH_VECTOR_KEY]
        )

        for key in self.get_prepared_statement_labels(inference_helper_columns):
            if _logger.isEnabledFor(logging.DEBUG):
                _logger.debug(f"Parametrize prepared statements for key {key}")
            self._parametrised_prepared_statements[key] = (
                self._parametrize_prepared_statements(
                    self.prepared_statements[key], batch=key.startswith("batch")
                )
            )

    def init_parametrize_and_serving_utils(
        self,
        prepared_statements: List[ServingPreparedStatement],
    ) -> None:
        if _logger.isEnabledFor(logging.DEBUG):
            _logger.debug(
                "Initializing parametrize and serving utils property using %s",
                json.dumps(prepared_statements, default=lambda x: x.__dict__, indent=2),
            )
        self.prefix_by_serving_index = {
            statement.prepared_statement_index: statement.prefix
            for statement in prepared_statements
        }
        self._feature_name_order_by_psp = {
            statement.prepared_statement_index: dict(
                [
                    (param.name, param.index)
                    for param in statement.prepared_statement_parameters
                ]
            )
            for statement in prepared_statements
        }
        if _logger.isEnabledFor(logging.DEBUG):
            _logger.debug("Build serving keys by PreparedStatementParameter.index")
        for sk in self._serving_keys:
            self.serving_key_by_serving_index[sk.join_index] = (
                self.serving_key_by_serving_index.get(sk.join_index, []) + [sk]
            )
        if _logger.isEnabledFor(logging.DEBUG):
            _logger.debug("Sort serving keys by PreparedStatementParameter.index")
        for join_index in self.serving_key_by_serving_index:
            # feature_name_order_by_psp do not include the join index when the joint feature only contains label only
            # But _serving_key_by_serving_index include the index when the join_index is 0 (left side)
            if join_index in self._feature_name_order_by_psp:
                self.serving_key_by_serving_index[join_index] = sorted(
                    self.serving_key_by_serving_index[join_index],
                    key=lambda _sk,
                    join_index=join_index: self.feature_name_order_by_psp[
                        join_index
                    ].get(_sk.feature_name, 0),
                )

    def _parametrize_prepared_statements(
        self,
        prepared_statements: List[ServingPreparedStatement],
        batch: bool,
    ) -> Dict[int, sql.text]:
        prepared_statements_dict = {}
        for prepared_statement in prepared_statements:
            if prepared_statement.feature_group_id in self._skip_fg_ids:
                continue
            query_online = str(prepared_statement.query_online).replace("\n", " ")

            if not batch:
                for param in prepared_statement.prepared_statement_parameters:
                    query_online = self._parametrize_query(param.name, query_online)
                query_online = sql.text(query_online)
            else:
                query_online = self._parametrize_query("batch_ids", query_online)
                query_online = sql.text(query_online)
                query_online = query_online.bindparams(
                    batch_ids=bindparam("batch_ids", expanding=True)
                )

            prepared_statements_dict[prepared_statement.prepared_statement_index] = (
                query_online
            )

        return prepared_statements_dict

    def init_async_mysql_connection(self, options=None):
        assert self._prepared_statements.get(self.SINGLE_VECTOR_KEY) is not None, (
            "Prepared statements are not initialized. "
            "Please call `init_prepared_statement` method first."
        )
        if _logger.isEnabledFor(logging.DEBUG):
            _logger.debug(
                "Fetching storage connector for sql connection to Online Feature Store."
            )
        self._online_connector = self._storage_connector_api.get_online_connector(
            self._feature_store_id
        )
        self._connection_options = options
        self._hostname = util.get_host_name() if self._external else None

        if util.is_runtime_notebook():
            if _logger.isEnabledFor(logging.DEBUG):
                _logger.debug("Running in Jupyter notebook, applying nest_asyncio")
            import nest_asyncio

            nest_asyncio.apply()
        else:
            if _logger.isEnabledFor(logging.DEBUG):
                _logger.debug("Running in python script. Not applying nest_asyncio")

    def get_single_feature_vector(self, entry: Dict[str, Any]) -> Dict[str, Any]:
        """Retrieve single vector with parallel queries using aiomysql engine."""
        return self._single_vector_result(
            entry, self.parametrised_prepared_statements[self.SINGLE_VECTOR_KEY]
        )

    def get_batch_feature_vectors(
        self, entries: List[Dict[str, Any]]
    ) -> List[Dict[str, Any]]:
        """Retrieve batch vector with parallel queries using aiomysql engine."""
        return self._batch_vector_results(
            entries, self.parametrised_prepared_statements[self.BATCH_VECTOR_KEY]
        )

    def get_inference_helper_vector(self, entry: Dict[str, Any]) -> Dict[str, Any]:
        """Retrieve single vector with parallel queries using aiomysql engine."""
        return self._single_vector_result(
            entry, self.parametrised_prepared_statements[self.SINGLE_HELPER_KEY]
        )

    def get_batch_inference_helper_vectors(
        self, entries: List[Dict[str, Any]]
    ) -> List[Dict[str, Any]]:
        """Retrieve batch vector with parallel queries using aiomysql engine."""
        return self._batch_vector_results(
            entries, self.parametrised_prepared_statements[self.BATCH_HELPER_KEY]
        )

    def _single_vector_result(
        self, entry: Dict[str, Any], prepared_statement_objects: Dict[int, sql.text]
    ) -> Dict[str, Any]:
        """Retrieve single vector with parallel queries using aiomysql engine."""

        if all([isinstance(val, list) for val in entry.values()]):
            raise ValueError(
                "Entry is expected to be single value per primary key. "
                "If you have already initialised prepared statements for single vector and now want to retrieve "
                "batch vector please reinitialise prepared statements with  "
                "`training_dataset.init_prepared_statement()` "
                "or `feature_view.init_serving()`"
            )
        # Initialize the set of values
        serving_vector = {}
        bind_entries = {}
        prepared_statement_execution = {}
        for prepared_statement_index in prepared_statement_objects:
            pk_entry = {}
            next_statement = False
            for sk in self.serving_key_by_serving_index[prepared_statement_index]:
                if sk.required_serving_key not in entry.keys():
                    # Check if there is any entry matched with feature name.
                    if sk.feature_name in entry.keys():
                        pk_entry[sk.feature_name] = entry[sk.feature_name]
                    else:
                        # User did not provide the necessary serving keys, we expect they have
                        # provided the necessary features as passed_features.
                        # We are going to check later if this is true
                        next_statement = True
                        break
                else:
                    pk_entry[sk.feature_name] = entry[sk.required_serving_key]
            if next_statement:
                continue
            bind_entries[prepared_statement_index] = pk_entry
            prepared_statement_execution[prepared_statement_index] = (
                prepared_statement_objects[prepared_statement_index]
            )

        # run all the prepared statements in parallel using aiomysql engine
        if _logger.isEnabledFor(logging.DEBUG):
            _logger.debug(
                f"Executing prepared statements for serving vector with entries: {bind_entries}"
            )
        loop = self._get_or_create_event_loop()
        results_dict = loop.run_until_complete(
            self._execute_prep_statements(prepared_statement_execution, bind_entries)
        )
        if _logger.isEnabledFor(logging.DEBUG):
            _logger.debug(f"Retrieved feature vectors: {results_dict}")
            _logger.debug("Constructing serving vector from results")
        for key in results_dict:
            for row in results_dict[key]:
                if _logger.isEnabledFor(logging.DEBUG):
                    _logger.debug(f"Processing row: {row} for prepared statement {key}")
                result_dict = dict(row)
                serving_vector.update(result_dict)

        return serving_vector

    def _batch_vector_results(
        self,
        entries: List[Dict[str, Any]],
        prepared_statement_objects: Dict[int, sql.text],
    ):
        """Execute prepared statements in parallel using aiomysql engine."""
        if _logger.isEnabledFor(logging.DEBUG):
            _logger.debug(
                f"Starting batch vector retrieval for {len(entries)} entries via aiomysql engine."
            )
        # create dict object that will have of order of the vector as key and values as
        # vector itself to stitch them correctly if there are multiple feature groups involved. At this point we
        # expect that backend will return correctly ordered vectors.
        batch_results = [{} for _ in range(len(entries))]
        entry_values = {}
        serving_keys_all_fg = []
        prepared_stmts_to_execute = {}
        # construct the list of entry values for binding to query
        if _logger.isEnabledFor(logging.DEBUG):
            _logger.debug(
                f"Parametrize prepared statements with entry values: {entries}"
            )
        for prepared_statement_index in prepared_statement_objects:
            # prepared_statement_index include fg with label only
            # But _serving_key_by_serving_index include the index when the join_index is 0 (left side)
            if prepared_statement_index not in self._serving_key_by_serving_index:
                continue

            prepared_stmts_to_execute[prepared_statement_index] = (
                prepared_statement_objects[prepared_statement_index]
            )
            entry_values_tuples = list(
                map(
                    lambda e, prepared_statement_index=prepared_statement_index: tuple(
                        [
                            (
                                e.get(sk.required_serving_key)
                                # Check if there is any entry matched with feature name,
                                # if the required serving key is not provided.
                                or e.get(sk.feature_name)
                            )
                            for sk in self.serving_key_by_serving_index[
                                prepared_statement_index
                            ]
                        ]
                    ),
                    entries,
                )
            )
            if _logger.isEnabledFor(logging.DEBUG):
                _logger.debug(
                    f"Prepared statement {prepared_statement_index} with entries: {entry_values_tuples}"
                )
            entry_values[prepared_statement_index] = {"batch_ids": entry_values_tuples}

        if _logger.isEnabledFor(logging.DEBUG):
            _logger.debug(
                f"Executing prepared statements for batch vector with entries: {entry_values}"
            )
        # run all the prepared statements in parallel using aiomysql engine
        loop = self._get_or_create_event_loop()
        parallel_results = loop.run_until_complete(
            self._execute_prep_statements(prepared_stmts_to_execute, entry_values)
        )

        if _logger.isEnabledFor(logging.DEBUG):
            _logger.debug(
                f"Retrieved feature vectors: {parallel_results}, stitching them."
            )
        # construct the results
        for prepared_statement_index in prepared_stmts_to_execute:
            statement_results = {}
            serving_keys = self.serving_key_by_serving_index[prepared_statement_index]
            serving_keys_all_fg += serving_keys
            prefix_features = [
                (self.prefix_by_serving_index[prepared_statement_index] or "")
                + sk.feature_name
                for sk in self.serving_key_by_serving_index[prepared_statement_index]
            ]
            if _logger.isEnabledFor(logging.DEBUG):
                _logger.debug(
                    f"Use prefix from prepare statement because prefix from serving key is collision adjusted {prefix_features}."
                )
                _logger.debug("iterate over results by index of the prepared statement")
            for row in parallel_results[prepared_statement_index]:
                if _logger.isEnabledFor(logging.DEBUG):
                    _logger.debug(f"Processing row: {row}")
                row_dict = dict(row)
                # can primary key be complex feature? No, not supported.
                result_dict = row_dict
                if _logger.isEnabledFor(logging.DEBUG):
                    _logger.debug(
                        f"Add result to statement results: {self._get_result_key(prefix_features, row_dict)} : {result_dict}"
                    )
                statement_results[self._get_result_key(prefix_features, row_dict)] = (
                    result_dict
                )
            if _logger.isEnabledFor(logging.DEBUG):
                _logger.debug(
                    f"Add partial results to batch results: {statement_results}"
                )
            for i, entry in enumerate(entries):
                if _logger.isEnabledFor(logging.DEBUG):
                    _logger.debug(
                        "Processing entry %s : %s",
                        entry,
                        statement_results.get(
                            self._get_result_key_serving_key(serving_keys, entry), {}
                        ),
                    )
                batch_results[i].update(
                    statement_results.get(
                        self._get_result_key_serving_key(serving_keys, entry), {}
                    )
                )
        return batch_results, serving_keys_all_fg

    def _get_or_create_event_loop(self):
        try:
            if _logger.isEnabledFor(logging.DEBUG):
                _logger.debug("Acquiring or starting event loop for async engine.")
            loop = asyncio.get_event_loop()
            asyncio.set_event_loop(loop)
        except RuntimeError as ex:
            if "There is no current event loop in thread" in str(ex):
                if _logger.isEnabledFor(logging.DEBUG):
                    _logger.debug(
                        "No existing running event loop. Creating new event loop."
                    )
                loop = asyncio.new_event_loop()
                asyncio.set_event_loop(loop)
        return loop

    def refresh_mysql_connection(self):
        if _logger.isEnabledFor(logging.DEBUG):
            _logger.debug("Refreshing MySQL connection.")
        try:
            if _logger.isEnabledFor(logging.DEBUG):
                _logger.debug("Checking if the connection is still alive.")
            with self._prepared_statement_engine.connect():
                # This will raise an exception if the connection is closed
                pass
        except exc.OperationalError:
            if _logger.isEnabledFor(logging.DEBUG):
                _logger.debug("Connection is closed, re-establishing connection.")
            self._set_mysql_connection()

    def _make_preview_statement(self, statement, n):
        return text(statement.text[: statement.text.find(" WHERE ")] + f" LIMIT {n}")

    def _set_mysql_connection(self, options=None):
        if _logger.isEnabledFor(logging.DEBUG):
            _logger.debug(
                "Retrieve MySQL connection details from the online storage connector."
            )
        online_conn = self._storage_connector_api.get_online_connector(
            self._feature_store_id
        )
        if _logger.isEnabledFor(logging.DEBUG):
            _logger.debug(
                f"Creating MySQL {'external' if self.external is True else ''}engine with options: {options}."
            )
        self._prepared_statement_engine = util.create_mysql_engine(
            online_conn, self._external, options=options
        )

    @staticmethod
    def _parametrize_query(name: str, query_online: str) -> str:
        # Now we have ordered pk_names, iterate over it and replace `?` with `:feature_name` one by one.
        # Regex `"^(.*?)\?"` will identify 1st occurrence of `?` in the sql string and replace it with provided
        # feature name, e.g. `:feature_name`:. As we iteratively update `query_online` we are always aiming to
        # replace 1st occurrence of `?`. This approach can only work if primary key names are sorted properly.
        # Regex `"^(.*?)\?"`:
        # `^` - asserts position at start of a line
        # `.*?` - matches any character (except for line terminators). `*?` Quantifier —
        # Matches between zero and unlimited times, expanding until needed, i.e 1st occurrence of `\?`
        # character.
        if _logger.isEnabledFor(logging.DEBUG):
            _logger.debug(f"Parametrizing name {name} in query {query_online}")
        return re.sub(
            r"^(.*?)\?",
            r"\1:" + name,
            query_online,
        )

    @staticmethod
    def _get_result_key(
        primary_keys: List[str], result_dict: Dict[str, str]
    ) -> Tuple[str]:
        if _logger.isEnabledFor(logging.DEBUG):
            _logger.debug(
                f"Get result key {primary_keys} from result dict {result_dict}"
            )
        result_key = []
        for pk in primary_keys:
            result_key.append(result_dict.get(pk))
        return tuple(result_key)

    @staticmethod
    def _get_result_key_serving_key(
        serving_keys: List["ServingKey"], result_dict: Dict[str, Dict[str, Any]]
    ) -> Tuple[str]:
        if _logger.isEnabledFor(logging.DEBUG):
            _logger.debug(
                f"Get result key serving key {serving_keys} from result dict {result_dict}"
            )
        result_key = []
        for sk in serving_keys:
            if _logger.isEnabledFor(logging.DEBUG):
                _logger.debug(
                    f"Get result key for serving key {sk.required_serving_key} or {sk.feature_name}"
                )
            result_key.append(
                result_dict.get(sk.required_serving_key)
                or result_dict.get(sk.feature_name)
            )
        if _logger.isEnabledFor(logging.DEBUG):
            _logger.debug(f"Result key: {result_key}")
        return tuple(result_key)

    @staticmethod
    def get_prepared_statement_labels(
        with_inference_helper_column: bool = False,
    ) -> List[str]:
        if with_inference_helper_column:
            return [
                OnlineStoreSqlClient.SINGLE_VECTOR_KEY,
                OnlineStoreSqlClient.BATCH_VECTOR_KEY,
                OnlineStoreSqlClient.SINGLE_HELPER_KEY,
                OnlineStoreSqlClient.BATCH_HELPER_KEY,
            ]
        return [
            OnlineStoreSqlClient.SINGLE_VECTOR_KEY,
            OnlineStoreSqlClient.BATCH_VECTOR_KEY,
        ]

    async def _get_connection_pool(self, default_min_size: int) -> None:
        self._connection_pool = await util.create_async_engine(
            self._online_connector,
            self._external,
            default_min_size,
            options=self._connection_options,
            hostname=self._hostname,
        )

    async def _query_async_sql(self, stmt, bind_params):
        """Query prepared statement together with bind params using aiomysql connection pool"""
        if self._connection_pool is None:
            await self._get_connection_pool(
                len(self._prepared_statements[self.SINGLE_VECTOR_KEY])
            )
        async with self._connection_pool.acquire() as conn:
            # Execute the prepared statement
            if _logger.isEnabledFor(logging.DEBUG):
                _logger.debug(
                    f"Executing prepared statement: {stmt} with bind params: {bind_params}"
                )
            cursor = await conn.execute(stmt, bind_params)
            # Fetch the result
            if _logger.isEnabledFor(logging.DEBUG):
                _logger.debug("Waiting for resultset.")
            resultset = await cursor.fetchall()
            if _logger.isEnabledFor(logging.DEBUG):
                _logger.debug(f"Retrieved resultset: {resultset}. Closing cursor.")
            await cursor.close()

        return resultset

    async def _execute_prep_statements(
        self,
        prepared_statements: Dict[int, str],
        entries: Union[List[Dict[str, Any]], Dict[str, Any]],
    ):
        """Iterate over prepared statements to create async tasks
        and gather all tasks results for a given list of entries."""

        # validate if prepared_statements and entries have the same keys
        if prepared_statements.keys() != entries.keys():
            # iterate over prepared_statements and entries to find the missing key
            # remove missing keys from prepared_statements
            for key in list(prepared_statements.keys()):
                if key not in entries:
                    prepared_statements.pop(key)

        try:
            tasks = [
                asyncio.create_task(
                    self._query_async_sql(prepared_statements[key], entries[key]),
                    name="query_prep_statement_key" + str(key),
                )
                for key in prepared_statements
            ]
            # Run the queries in parallel using asyncio.gather
            results = await asyncio.gather(*tasks)
        except asyncio.CancelledError as e:
            if _logger.isEnabledFor(logging.ERROR):
                _logger.error(f"Failed executing prepared statements: {e}")
            raise e

        # Create a dict of results with the prepared statement index as key
        results_dict = {}
        for i, key in enumerate(prepared_statements):
            results_dict[key] = results[i]

        return results_dict

    @property
    def feature_store_id(self) -> int:
        return self._feature_store_id

    @property
    def prepared_statement_engine(self) -> Optional[Any]:
        """JDBC connection engine to retrieve connections to online features store from."""
        return self._prepared_statement_engine

    @prepared_statement_engine.setter
    def prepared_statement_engine(self, prepared_statement_engine: Any) -> None:
        self._prepared_statement_engine = prepared_statement_engine

    @property
    def prepared_statements(
        self,
    ) -> Dict[str, List[ServingPreparedStatement]]:
        """Contains up to 4 prepared statements for single and batch vector retrieval, and single or batch inference helpers.

        The keys are the labels for the prepared statements, and the values are dictionaries of prepared statements
        with the prepared statement index as the key.
        """
        return self._prepared_statements

    @prepared_statements.setter
    def prepared_statements(
        self,
        prepared_statements: Dict[str, List[ServingPreparedStatement]],
    ) -> None:
        self._prepared_statements = prepared_statements

    @property
    def parametrised_prepared_statements(
        self,
    ) -> Dict[str, Dict[int, sql.text]]:
        """The dict object of prepared_statements as values and keys as indices of positions in the query for
        selecting features from feature groups of the training dataset. Used for batch retrieval.
        """
        return self._parametrised_prepared_statements

    @parametrised_prepared_statements.setter
    def parametrised_prepared_statements(
        self,
        parametrised_prepared_statements: Dict[str, Dict[int, sql.text]],
    ) -> None:
        self._parametrised_prepared_statements = parametrised_prepared_statements

    @property
    def prefix_by_serving_index(self) -> Dict[int, str]:
        """The dict object of prefixes as values and keys as indices of positions in the query for
        selecting features from feature groups of the training dataset.
        """
        return self._prefix_by_serving_index

    @prefix_by_serving_index.setter
    def prefix_by_serving_index(self, prefix_by_serving_index: Dict[int, str]) -> None:
        if _logger.isEnabledFor(logging.DEBUG):
            _logger.debug(f"Setting prefix by serving index {prefix_by_serving_index}.")
        self._prefix_by_serving_index = prefix_by_serving_index

    @property
    def serving_key_by_serving_index(
        self,
    ) -> Dict[int, List[ServingKey]]:
        """The dict object of serving keys as values and keys as indices of positions in the query for
        selecting features from feature groups of the training dataset.
        """
        return self._serving_key_by_serving_index

    @property
    def feature_name_order_by_psp(self) -> Dict[int, Dict[str, int]]:
        """The dict object of feature names as values and keys as indices of positions in the query for
        selecting features from feature groups of the training dataset.
        """
        return self._feature_name_order_by_psp

    @property
    def skip_fg_ids(self) -> Set[int]:
        """The list of feature group ids to skip when retrieving feature vectors.

        The retrieval of Feature values stored in Feature Group with embedding is handled via a separate client
        as there are not stored in RonDB.
        """
        return self._skip_fg_ids

    @property
    def serving_keys(self) -> Set[ServingKey]:
        if len(self._serving_keys) > 0:
            return self._serving_keys

        if len(self.prepared_statements) == 0:
            raise ValueError(
                "Prepared statements are not initialized. Please call `init_prepared_statement` method first."
            )
        else:
            if _logger.isEnabledFor(logging.DEBUG):
                _logger.debug(
                    "Build serving keys from prepared statements ignoring prefix to ensure compatibility with older version."
                )
            self._serving_keys = util.build_serving_keys_from_prepared_statements(
                self.prepared_statements[
                    self.BATCH_VECTOR_KEY
                ],  # use batch to avoid issue with label_fg
                ignore_prefix=True,  # if serving_keys are not set it is because the feature view is anterior to 3.3, this ensures compatibility
            )
        return self._serving_keys

    @property
    def training_dataset_api(self) -> training_dataset_api.TrainingDatasetApi:
        return self._training_dataset_api

    @property
    def feature_view_api(self) -> feature_view_api.FeatureViewApi:
        return self._feature_view_api

    @property
    def storage_connector_api(self) -> storage_connector_api.StorageConnectorApi:
        return self._storage_connector_api

    @property
    def hostname(self) -> str:
        return self._hostname

    @property
    def connection_options(self) -> Dict[str, Any]:
        return self._connection_options

    @property
    def online_connector(self) -> storage_connector.StorageConnector:
        return self._online_connector

    @property
    def connection_pool(self) -> aiomysql.utils._ConnectionContextManager:
        return self._connection_pool



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import importlib.util


polars_not_installed_message = (
    "Polars package not found. "
    "If you want to use Polars with Hopsworks you can install the corresponding extras "
    """`pip install hopsworks[polars]` or `pip install "hopsworks[polars]"` if using zsh. """
    "You can also install polars directly in your environment e.g `pip install polars`. "
    "You will need to restart your kernel if applicable."
)

HAS_POLARS: bool = importlib.util.find_spec("polars") is not None



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import os
from pathlib import Path

import requests
from hsfs.client import auth, base


try:
    import jks
except ImportError:
    pass


class Client(base.Client):
    REQUESTS_VERIFY = "REQUESTS_VERIFY"
    DOMAIN_CA_TRUSTSTORE_PEM = "DOMAIN_CA_TRUSTSTORE_PEM"
    PROJECT_ID = "HOPSWORKS_PROJECT_ID"
    PROJECT_NAME = "HOPSWORKS_PROJECT_NAME"
    HADOOP_USER_NAME = "HADOOP_USER_NAME"
    MATERIAL_DIRECTORY = "MATERIAL_DIRECTORY"
    HDFS_USER = "HDFS_USER"
    T_CERTIFICATE = "t_certificate"
    K_CERTIFICATE = "k_certificate"
    TRUSTSTORE_SUFFIX = "__tstore.jks"
    KEYSTORE_SUFFIX = "__kstore.jks"
    PEM_CA_CHAIN = "ca_chain.pem"
    CERT_KEY_SUFFIX = "__cert.key"
    MATERIAL_PWD = "material_passwd"
    SECRETS_DIR = "SECRETS_DIR"

    def __init__(self):
        """Initializes a client being run from a job/notebook directly on Hopsworks."""
        self._base_url = self._get_hopsworks_rest_endpoint()
        self._host, self._port = self._get_host_port_pair()
        self._secrets_dir = (
            os.environ[self.SECRETS_DIR] if self.SECRETS_DIR in os.environ else ""
        )
        self._cert_key = self._get_cert_pw()
        trust_store_path = self._get_trust_store_path()
        hostname_verification = (
            os.environ[self.REQUESTS_VERIFY]
            if self.REQUESTS_VERIFY in os.environ
            else "true"
        )
        self._project_id = os.environ[self.PROJECT_ID]
        self._project_name = self._project_name()
        try:
            self._auth = auth.BearerAuth(self._read_jwt())
        except FileNotFoundError:
            self._auth = auth.ApiKeyAuth(self._read_apikey())
        self._verify = self._get_verify(hostname_verification, trust_store_path)
        self._session = requests.session()

        self._connected = True

        credentials = self._get_credentials(self._project_id)

        self._write_pem_file(credentials["caChain"], self._get_ca_chain_path())
        self._write_pem_file(credentials["clientCert"], self._get_client_cert_path())
        self._write_pem_file(credentials["clientKey"], self._get_client_key_path())

    def _get_hopsworks_rest_endpoint(self):
        """Get the hopsworks REST endpoint for making requests to the REST API."""
        return os.environ[self.REST_ENDPOINT]

    def _get_trust_store_path(self):
        """Convert truststore from jks to pem and return the location"""
        ca_chain_path = Path(self.PEM_CA_CHAIN)
        if not ca_chain_path.exists():
            ks = jks.KeyStore.load(
                self._get_jks_key_store_path(), self._cert_key, try_decrypt_keys=True
            )
            ts = jks.KeyStore.load(
                self._get_jks_trust_store_path(), self._cert_key, try_decrypt_keys=True
            )
            self._write_ca_chain(
                ks,
                ts,
                ca_chain_path,
            )
        return str(ca_chain_path)

    def _get_ca_chain_path(self) -> str:
        return os.path.join("/tmp", "ca_chain.pem")

    def _get_client_cert_path(self) -> str:
        return os.path.join("/tmp", "client_cert.pem")

    def _get_client_key_path(self) -> str:
        return os.path.join("/tmp", "client_key.pem")

    def _get_jks_trust_store_path(self):
        """
        Get truststore location

        Returns:
             truststore location
        """
        t_certificate = Path(self.T_CERTIFICATE)
        if t_certificate.exists():
            return str(t_certificate)
        else:
            username = os.environ[self.HADOOP_USER_NAME]
            material_directory = Path(os.environ[self.MATERIAL_DIRECTORY])
            return str(material_directory.joinpath(username + self.TRUSTSTORE_SUFFIX))

    def _get_jks_key_store_path(self):
        """
        Get keystore location

        Returns:
             keystore location
        """
        k_certificate = Path(self.K_CERTIFICATE)
        if k_certificate.exists():
            return str(k_certificate)
        else:
            username = os.environ[self.HADOOP_USER_NAME]
            material_directory = Path(os.environ[self.MATERIAL_DIRECTORY])
            return str(material_directory.joinpath(username + self.KEYSTORE_SUFFIX))

    def _project_name(self):
        try:
            return os.environ[self.PROJECT_NAME]
        except KeyError:
            pass

        hops_user = self._project_user()
        hops_user_split = hops_user.split(
            "__"
        )  # project users have username project__user
        project = hops_user_split[0]
        return project

    def _project_user(self):
        try:
            hops_user = os.environ[self.HADOOP_USER_NAME]
        except KeyError:
            hops_user = os.environ[self.HDFS_USER]
        return hops_user

    def _get_cert_pw(self):
        """
        Get keystore password from local container

        Returns:
            Certificate password
        """
        pwd_path = Path(self.MATERIAL_PWD)
        if not pwd_path.exists():
            username = os.environ[self.HADOOP_USER_NAME]
            material_directory = Path(os.environ[self.MATERIAL_DIRECTORY])
            pwd_path = material_directory.joinpath(username + self.CERT_KEY_SUFFIX)

        with pwd_path.open() as f:
            return f.read()

    def replace_public_host(self, url):
        """replace hostname to public hostname set in HOPSWORKS_PUBLIC_HOST"""
        ui_url = url._replace(netloc=os.environ[self.HOPSWORKS_PUBLIC_HOST])
        return ui_url

    def _is_external(self):
        return False

    @property
    def host(self):
        return self._host



#
#   Copyright 2021 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import json

from hsfs import util


class Code:
    def __init__(
        self,
        commit_time,
        application_id,
        content=None,
        feature_group_commit_id=None,
        href=None,
        expand=None,
        items=None,
        count=None,
        type=None,
        **kwargs,
    ):
        self._commit_time = commit_time
        self._application_id = application_id

    def to_dict(self):
        return {
            "commitTime": self._commit_time,
            "applicationId": self._application_id,
        }

    def json(self):
        return json.dumps(self, cls=util.FeatureStoreEncoder)

    @property
    def commit_time(self):
        return self._commit_time

    @property
    def application_id(self):
        return self._application_id

    @property
    def feature_group_commit_id(self):
        return self._feature_group_commit_id

    @property
    def content(self):
        return self._content



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import re
from datetime import datetime, timedelta
from typing import List, Optional, Tuple, TypeVar, Union

from hsfs import feature_group, feature_view, util
from hsfs.client.exceptions import RestAPIError
from hsfs.core import monitoring_window_config as mwc
from hsfs.core import statistics_engine
from hsfs.core.feature_descriptive_statistics import FeatureDescriptiveStatistics
from hsfs.training_dataset_split import TrainingDatasetSplit


class MonitoringWindowConfigEngine:
    _MAX_TIME_RANGE_LENGTH = 12

    def __init__(self, **kwargs):
        # No need to initialize anything
        pass

    def _init_statistics_engine(self, feature_store_id: int, entity_type: str):
        self._statistics_engine = statistics_engine.StatisticsEngine(
            feature_store_id=feature_store_id,
            entity_type=entity_type,
        )

    def validate_monitoring_window_config(
        self,
        time_offset: Optional[str] = None,
        window_length: Optional[str] = None,
        training_dataset_version: Optional[int] = None,
        specific_value: Optional[Union[int, float]] = None,
        row_percentage: Optional[float] = None,
    ) -> "mwc.WindowConfigType":
        if isinstance(specific_value, int) or isinstance(specific_value, float):
            if any(
                [
                    time_offset is not None,
                    window_length is not None,
                    training_dataset_version is not None,
                    row_percentage is not None,
                ]
            ):
                raise ValueError(
                    "If specific_value is set, no other parameter can be set."
                )
            return mwc.WindowConfigType.SPECIFIC_VALUE

        if isinstance(training_dataset_version, int):
            if any(
                [
                    time_offset is not None,
                    window_length is not None,
                    row_percentage is not None,
                ]
            ):
                raise ValueError(
                    "If training_dataset_version is set, no other parameter can be set."
                )
            return mwc.WindowConfigType.TRAINING_DATASET

        if isinstance(time_offset, str):
            return mwc.WindowConfigType.ROLLING_TIME

        if isinstance(window_length, str):
            raise ValueError("window_length can only be set if time_offset is set.")

        return mwc.WindowConfigType.ALL_TIME

    def build_monitoring_window_config(
        self,
        id: Optional[int] = None,
        window_config_type: Optional[Union["mwc.WindowConfigType", str]] = None,
        time_offset: Optional[str] = None,
        window_length: Optional[str] = None,
        training_dataset_version: Optional[int] = None,
        specific_value: Optional[Union[int, float]] = None,
        row_percentage: Optional[float] = None,
    ) -> "mwc.MonitoringWindowConfig":
        """Builds a monitoring window config.

        Args:
            window_config_type: str, required
                Type of the window config, can be either
                `ROLLING_TIME`,`SPECIFIC_VALUE`,`TRAINING_DATASET`.
            time_offset: str, optional
                monitoring window start time is computed as "now - time_offset".
            window_length: str, optional
                monitoring window end time is computed as
                    "now - time_offset + window_length".
            training_dataset_version: int, optional
                Specific id of an entity that has fixed statistics.
            specific_value: float, optional
                Specific value instead of a statistics computed on data.
            row_percentage: float, optional
                Percentage of rows to be used for statistics computation.
            id: int, optional
                Id of the monitoring window config in hopsworks.

        Returns:
            MonitoringWindowConfig The monitoring window configuration.
        """

        detected_window_config_type = self.validate_monitoring_window_config(
            time_offset=time_offset,
            window_length=window_length,
            training_dataset_version=training_dataset_version,
            specific_value=specific_value,
            row_percentage=row_percentage,
        )

        if (
            isinstance(window_config_type, str)
            and window_config_type != detected_window_config_type
        ):
            raise ValueError(
                "The window_config_type does not match the parameters set."
            )

        if (
            window_config_type
            in [mwc.WindowConfigType.ROLLING_TIME, mwc.WindowConfigType.ALL_TIME]
            and row_percentage is None
        ):
            row_percentage = 1.0

        return mwc.MonitoringWindowConfig(
            id=id,
            window_config_type=detected_window_config_type,
            time_offset=time_offset,
            window_length=window_length,
            training_dataset_version=training_dataset_version,
            specific_value=specific_value,
            row_percentage=row_percentage,
        )

    def time_range_str_to_time_delta(
        self, time_range: str, field_name: Optional[str] = "time_offset"
    ) -> timedelta:
        # sanitize input
        value_error_message = f"Invalid {field_name} format: {time_range}. Use format: 1w2d3h for 1 week, 2 days and 3 hours."
        if (
            len(time_range) > self._MAX_TIME_RANGE_LENGTH
            or re.search(r"([^dwh\d]+)", time_range) is not None
        ):
            raise ValueError(value_error_message)

        matches = re.search(
            # r"^(?!$)(?:.*(?P<week>\d+)w)?(?:.*(?P<day>\d+)d)?(?:.*(?P<hour>\d+)h)?$",
            r"(?:(?P<week>\d+w)()|(?P<day>\d+d)()|(?P<hour>\d+h)())+",
            time_range,
        )
        if matches is None:
            raise ValueError(value_error_message)

        weeks = (
            int(matches.group("week").replace("w", ""))
            if matches.group("week") is not None
            else 0
        )
        days = (
            int(matches.group("day").replace("d", ""))
            if matches.group("day") is not None
            else 0
        )
        hours = (
            int(matches.group("hour").replace("h", ""))
            if matches.group("hour") is not None
            else 0
        )

        return timedelta(weeks=weeks, days=days, hours=hours)

    def get_window_start_end_times(
        self,
        monitoring_window_config: "mwc.MonitoringWindowConfig",
    ) -> Tuple[Optional[int], int]:
        end_time = datetime.now()
        if monitoring_window_config.window_config_type not in [
            mwc.WindowConfigType.ROLLING_TIME,
            mwc.WindowConfigType.ALL_TIME,
        ]:
            return (
                None,
                self.round_and_convert_event_time(event_time=end_time),
            )

        if monitoring_window_config.time_offset is not None:
            time_offset = self.time_range_str_to_time_delta(
                monitoring_window_config.time_offset
            )
            start_time = datetime.now() - time_offset
        else:
            # case where time_offset is None and window_length is None
            return (
                None,
                self.round_and_convert_event_time(event_time=end_time),
            )

        if monitoring_window_config.window_length is not None:
            window_length = self.time_range_str_to_time_delta(
                monitoring_window_config.window_length
            )
            end_time = (
                start_time + window_length
                if start_time + window_length < end_time
                else end_time
            )

        return (
            self.round_and_convert_event_time(event_time=start_time),
            self.round_and_convert_event_time(event_time=end_time),
        )

    def run_single_window_monitoring(
        self,
        entity: Union["feature_group.FeatureGroup", "feature_view.FeatureView"],
        monitoring_window_config: "mwc.MonitoringWindowConfig",
        feature_name: Optional[str] = None,
    ) -> List[FeatureDescriptiveStatistics]:
        """Fetch the entity data based on monitoring window configuration and compute statistics.

        Args:
            entity: FeatureStore: Feature store to fetch the entity to monitor.
            monitoring_window_config: MonitoringWindowConfig: Monitoring window config.
            feature_name: str: Name of the feature to monitor.

        Returns:
            [FeatureDescriptiveStatistics, List[FeatureDescriptiveStatitics]]: List of Descriptive statistics.
        """
        self._init_statistics_engine(entity._feature_store_id, entity.ENTITY_TYPE)
        (
            start_time,
            end_time,
        ) = self.get_window_start_end_times(
            monitoring_window_config=monitoring_window_config,
        )
        if (
            monitoring_window_config.window_config_type
            == mwc.WindowConfigType.TRAINING_DATASET
            and isinstance(entity, feature_view.FeatureView)
        ):
            td_meta = entity._feature_view_engine._get_training_dataset_metadata(
                entity,
                training_dataset_version=monitoring_window_config.training_dataset_version,
            )
            before_transformation = (
                feature_name is not None
                and td_meta.transformation_functions is not None
                and feature_name in td_meta.transformation_functions.keys()
            )
            registered_stats = entity.get_training_dataset_statistics(
                training_dataset_version=monitoring_window_config.training_dataset_version,
                before_transformation=before_transformation,
                feature_names=[feature_name] if feature_name is not None else None,
            )
            if (
                registered_stats.feature_descriptive_statistics is None
                and registered_stats.split_statistics is not None
            ):
                # if td splits, we use the train set statistics
                for split in registered_stats.split_statistics:
                    if split.name == TrainingDatasetSplit.TRAIN:
                        registered_stats = split
        else:
            # Check if statistics already exists
            registered_stats = self._statistics_engine.get_by_time_window(
                metadata_instance=entity,
                start_commit_time=start_time,
                end_commit_time=end_time,
                feature_names=[feature_name] if feature_name is not None else None,
                row_percentage=monitoring_window_config.row_percentage,
            )

        if registered_stats is None:  # if statistics don't exist
            # Fetch the actual data for which to compute statistics based on row_percentage and time window
            entity_feature_df = self.fetch_entity_data_in_monitoring_window(
                entity=entity,
                feature_name=feature_name,
                start_time=start_time,
                end_time=end_time,
                row_percentage=monitoring_window_config.row_percentage,
            )

            # Compute statistics on the feature dataframe
            registered_stats = (
                self._statistics_engine.compute_and_save_monitoring_statistics(
                    entity,
                    feature_dataframe=entity_feature_df,
                    window_start_commit_time=start_time,
                    window_end_commit_time=end_time,
                    row_percentage=monitoring_window_config.row_percentage,
                    feature_name=feature_name,
                )
            )

        assert (
            registered_stats.feature_descriptive_statistics is not None
        ), "statistics should contain the feature descriptive statistics"

        return registered_stats.feature_descriptive_statistics

    def fetch_entity_data_in_monitoring_window(
        self,
        entity: Union["feature_group.FeatureGroup", "feature_view.FeatureView"],
        start_time: Optional[int],
        end_time: Optional[int],
        row_percentage: float,
        feature_name: Optional[str] = None,
    ) -> TypeVar("pyspark.sql.DataFrame"):
        """Fetch the entity data based on time window and row percentage.

        Args:
            entity: Union[FeatureGroup, FeatureView]: Entity to monitor.
            feature_name: str: Name of the feature to monitor.
            start_time: int: Window start commit or event time
            end_time: int: Window end commit or event time
            row_percentage: fraction of rows to include [0, 1.0]

        Returns:
            `pyspark.sql.DataFrame`. A Spark DataFrame with the entity data
        """
        try:
            if isinstance(entity, feature_group.FeatureGroup):
                entity_df = self.fetch_feature_group_data(
                    entity=entity,
                    feature_name=feature_name,
                    start_time=start_time,
                    end_time=end_time,
                )
            else:
                entity_df = self.fetch_feature_view_data(
                    entity=entity,
                    feature_name=feature_name,
                    start_time=start_time,
                    end_time=end_time,
                )

            if row_percentage < 1.0:
                entity_df = entity_df.sample(fraction=row_percentage)

        except RestAPIError as e:
            if (
                e.response.json().get("errorCode", "") == 270118
                and e.response.status_code == 404
            ):
                # creating an empty Spark DataFrame requires a valid schema [SchemaType(name, type)]
                # and takes unnecessary time and resources. We can return an empty pandas dataframe instead
                # because the computation of statistics will be discarded in statistics_engine (len(df.head()) == 0)
                import pandas as pd

                return pd.DataFrame(columns=[feat.name for feat in entity.schema])
            else:
                raise e

        return entity_df

    def fetch_feature_view_data(
        self,
        entity: "feature_view.FeatureView",
        feature_name: Optional[str] = None,
        start_time: Optional[int] = None,
        end_time: Optional[int] = None,
    ) -> TypeVar("pyspark.sql.DataFrame"):
        """Fetch the feature view data based on time window and row percentage.

        Args:
            entity: FeatureView: Feature view to monitor.
            feature_name: str: Name of the feature to monitor.
            start_time: int: Window start commit or event time.
            end_time: int: Window end commit or event time.

        Returns:
            `pyspark.sql.DataFrame`. A Spark DataFrame with the entity data
        """

        # TODO: This fails for FV on non-time-travel FGs.
        entity_df = entity.query.as_of(
            exclude_until=start_time, wallclock_time=end_time
        ).read()

        if feature_name:
            entity_df = entity_df.select(feature_name)

        return entity_df

    def fetch_feature_group_data(
        self,
        entity: "feature_group.FeatureGroup",
        feature_name: Optional[str] = None,
        start_time: Optional[int] = None,
        end_time: Optional[int] = None,
    ) -> TypeVar("pyspark.sql.Dataframe"):
        """Fetch the feature group data based on time window.

        Args:
            entity: FeatureGroup: Feature group to monitor.
            feature_name: str: Name of the feature to monitor.
            start_time: int: Window start commit time.
            end_time: int: Window end commit time.
        """
        if feature_name:
            pre_df = entity.select(features=[feature_name])
        else:
            pre_df = entity

        full_df = pre_df.as_of(exclude_until=start_time, wallclock_time=end_time).read()

        return full_df

    def round_and_convert_event_time(self, event_time: datetime) -> Optional[int]:
        """Round event time to the latest hour and convert to timestamp.

        Args:
            event_time: datetime: Event time to round and convert.

        Returns:
            datetime: Rounded and converted event time.
        """
        return util.convert_event_time_to_timestamp(event_time)
